#include "bezierspline.h"

// Default to parametric and to showing control lines.
BezierSpline::BezierSpline(float m_dt) : m_dt(m_dt), m_bCasteljou(false), m_bShowLines(true)
{

	float startCJ = 0;
	float endCJ = 0;
	float startPara = 0;
	float endPara = 0;
	float CJtime = 0;
	float ParaTime= 0;
}

BezierSpline::~BezierSpline()
{
	Destroy();
}

void BezierSpline::Destroy()
{
	// delete memory taken up by our control point objects.

	for(int i = 0; i < (int)size(); i++)
		delete at(i);

	cout << "Bezier spline succesfully destroyed." << endl;
}

float& BezierSpline::dt()
{		
	return m_dt;
}

void BezierSpline::build()
{
	C = vector<Vector3>(size());

	for(int j = 0; j < (int)size(); j++) C[j] = Cj(j);

	// Update casteljau points.

	casPoints.clear();

	for(int i = 0; i < (int)size(); i++)
		casPoints.push_back(*at(i));
}

bool& BezierSpline::Casteljau()
{
	return m_bCasteljou;
}

bool& BezierSpline::ShowLines()
{
	return m_bShowLines;
}

/*

This is the method called to obtain a 3D point from the bezier spline, given a value t,
between 0 and 1 inclusive. The function uses the previously selected method to generate the spline.

*/

Vector3 BezierSpline::f(float t)
{		
	if(Casteljau())
	{
		// Size of coeficient vector doesnt match the number of points, so recalculate it.
		if(casPoints.size() != size()) build();

		// For the casteljau method, just copy the most current points to the temp vector.
		for(int i = 0; i < (int)size(); i++) casPoints[i] = *at(i);
	}

	// Call appropriate function.
	return Casteljau() ? fCasteljau(t, size()) : fPoly(t);

	//return fCasteljau(t, size());
}

//Used for simple "performance analysis" to see how fast the spline in drawn using parametric and Bezier.
void BezierSpline::Time()
{

	printf("\nTime taken for 10000 iterations with %d points\n", (int) size());
	m_bCasteljou = true;
	float startCJ = (float) GetTickCount();
	for(int i=0; i<10000;i++)
	{
		//float startCJ = (float) GetTickCount();
		f(0.5);
		//float endCJ = (float) GetTickCount();
	}
	float endCJ = (float) GetTickCount();
	CJtime = (endCJ - startCJ)/1000;
	printf("\ndeCasteljou = %.6f seconds\n", CJtime);

	m_bCasteljou = false;
	if((int)size() <= 10) 
	{

		float startPara = (float) GetTickCount();
		for(int i=0; i<10000; i++)
		{
			//float startPara = (float) GetTickCount();
			f(0.5);
			//float endPara = (float) GetTickCount();
		}
		float endPara = (float) GetTickCount();

		ParaTime = (endPara - startPara)/1000;		
		printf("Parametric = %.6f seconds\n", ParaTime  );
	}
	else printf("Parametric calculations unstable for more than 10 control points\n");
	//startCJ, endCJ, startPara, endPara = 0;
}

void BezierSpline::clear()
{
	C.clear();

	vector<GLObject*>::clear();
}

void BezierSpline::Render()
{		
	save();

	transform();

	glDisable(GL_LIGHTING);

	GLObject::Render();	// Takes care of transformations etc.

	// Outline.

	if(ShowLines())
	{
		glBegin(GL_LINE_STRIP);

		glColor3f(1, 0, 0);	// Red outline.

		for(int i = 0; i < (int)size(); i++)
		{
			glColor3f(1.0, 0, 0);
			glVertex3fv(at(i)->m_V);
		}

		glEnd();
	}

	// Curve.

	glBegin(GL_LINE_STRIP);

	glColor3f(0, 1, 0);	// Green outline.

	for(float t = 0; t < 1.0; t += m_dt)
		glVertex3fv(f(t).m_V);

	// Do the last vertex because we might not be exact with the above loop.
	glVertex3fv(f(1.0f).m_V);

	glEnd();

	restore();

	save();

	transform();

	// Control points.

	// Uncomment below to make the control points show through objects (terrain etc.)
	// glDisable(GL_DEPTH_TEST); // We want to see the control points if they are hidden.
	// glDisable(GL_LIGHTING);

	for(int i = 0; i < (int)size(); i++)
	{
		glColor3f(1.0, 0, 0);
		at(i)->Render();
	}

	restore();
}

// Basic factorial function.
int BezierSpline::fact(int j)
{
	int f = 1;

	for(int i = 2; i <= j; i++) f *= i;		// Simple factorial calculation.

	return f;
}

/*

The calculation of Cj coeficients can be found on Wikipedia.
The below function breaks up the calculation into terms and performs
products and sumations using for loops just as in the mathematical notation.

*/

Vector3 BezierSpline::Cj(int j)
{
	int		n = size()-1;
	int		p = 1;			// Product term
	Vector3 s;				// Sum term

	for(int m = 0; m < j; m++)
		p *= (n-m);

	for(int i=0; i <= j; i++)
		s += (powf((-1.0f), (float)(i+j))*(*at(i)))/(float)(fact(i)*fact(j-i));

	return (float)p*s;
}

/*

The basic parametric calculation is done on any number of points by expanding the basic eqaution to any order.
Instead of trying to join many fourth order curves, the equation is re-written for an nth order curve using binomial theorem etc.
This means no need to consider continuity conditions.

*/

Vector3 BezierSpline::fPoly(float t)
{
	if(C.size() != size()) build();	// Check for CJs update

	Vector3 r;
	int	n = size()-1;

	// The resulting point is a polynomial interpolation of the various Cj terms.
	
	for(int j = 0; j <= n; j++)
		r += powf(t, (float)j) * C[j];

	return r;
}

/*

The Casteljau method recursively calculates the final point by interpolating linearly at each step.
It takes a set of points and generates new points using a linear interpolation of adjacent points.
The new points will be one less than before and is recursively passed into
the same function until there is only one point remaining. That point is the answer.

*/

Vector3 BezierSpline::fCasteljau(float t, int j)	// Recursive bernstein
{
	// If there is only one point left, then its the answer!
	if(j == 1) return casPoints[0];

	for(int i = 0; i < j-1; i++)
		casPoints[i] = ((1-t)*casPoints[i] + t*casPoints[i+1]);		// Linear combination of adjacent points.

	// Points reduced by 1, restart calculation unless one point left.
	return fCasteljau(t, j-1);
}